import abc
from ..schema.base_schema import BlobValue as BlobValue, Device as Device, Network as Network, NumberValue as NumberValue, State as State, StringValue as StringValue, WappstoMethods as WappstoMethods, XmlValue as XmlValue
from _typeshed import Incomplete
from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable, Optional, Union
from uuid import UUID

class StatusID(str, Enum):
    IDLE: str
    SENDING: str
    SENDERROR: str
    SEND: str
    ERROR: str

class ServiceClass(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def subscribe_network_event(self, uuid: UUID, callback: Callable[[Network, WappstoMethods], None]): ...
    @abstractmethod
    def post_network(self, data) -> bool: ...
    @abstractmethod
    def put_network(self, uuid: UUID, data) -> bool: ...
    @abstractmethod
    def get_network(self, uuid: UUID) -> Union[Network, None]: ...
    @abstractmethod
    def delete_network(self, uuid: UUID) -> bool: ...
    @abstractmethod
    def subscribe_device_event(self, uuid: UUID, callback: Callable[[Device, WappstoMethods], None]): ...
    @abstractmethod
    def post_device(self, network_uuid: UUID, data: Device) -> bool: ...
    @abstractmethod
    def put_device(self, uuid: UUID, data: Device) -> bool: ...
    @abstractmethod
    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]: ...
    @abstractmethod
    def get_device(self, uuid: UUID) -> Union[Device, None]: ...
    @abstractmethod
    def delete_device(self, uuid: UUID) -> bool: ...
    ValueUnion: Incomplete
    @abstractmethod
    def subscribe_value_event(self, uuid: UUID, callback: Callable[[ValueUnion, WappstoMethods], None]): ...
    @abstractmethod
    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool: ...
    @abstractmethod
    def put_value(self, uuid: UUID, data: ValueUnion) -> bool: ...
    @abstractmethod
    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]: ...
    @abstractmethod
    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]: ...
    @abstractmethod
    def delete_value(self, uuid: UUID) -> bool: ...
    @abstractmethod
    def subscribe_state_event(self, uuid: UUID, callback: Callable[[State, WappstoMethods], None]): ...
    @abstractmethod
    def post_state(self, value_uuid: UUID, data: State) -> bool: ...
    @abstractmethod
    def put_state(self, uuid: UUID, data: State) -> bool: ...
    @abstractmethod
    def get_state(self, uuid: UUID) -> Union[State, None]: ...
    @abstractmethod
    def delete_state(self, uuid: UUID) -> bool: ...



# generated by datamodel-codegen:
#   filename:  schema.json
#   timestamp: 2021-07-08T11:31:02+00:00

# from __future__ import annotations

from datetime import datetime
from enum import Enum

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import conint
from pydantic import constr
from pydantic import Extra
from pydantic import Field
from pydantic import root_validator
from pydantic import UUID4


def timestamp_converter(dt: datetime) -> str:
    """
    Return The default timestamp used for Wappsto.

    The timestamp are always set to the UTC timezone.

    Returns:
        The UTC time string in ISO format.
    """
    return dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')


class WappstoMethods(str, Enum):
    """The methods that the Wappsto are using."""
    DELETE = "DELETE"
    PUT = "PUT"
    POST = "POST"
    GET = "GET"


class InclusionStatus(str, Enum):
    STATUS_DEVICE_INCLUDING = 'STATUS_DEVICE_INCLUDING'
    STATUS_DEVICE_INCLUSION_SUCCESS = 'STATUS_DEVICE_INCLUSION_SUCCESS'
    STATUS_DEVICE_INCLUSION_FAILURE = 'STATUS_DEVICE_INCLUSION_FAILURE'
    STATUS_DEVICE_REPORTING = 'STATUS_DEVICE_REPORTING'
    STATUS_DEVICE_REPORT_SUCCESS = 'STATUS_DEVICE_REPORT_SUCCESS'
    STATUS_DEVICE_REPORT_FAILURE = 'STATUS_DEVICE_REPORT_FAILURE'
    EXCLUDED = 'EXCLUDED'
    INCLUDED = 'INCLUDED'


class FirmwareStatus(str, Enum):
    UP_TO_DATE = 'UP_TO_DATE'
    UPDATE_AVAILABLE = 'UPDATE_AVAILABLE'
    UPLOADING = 'UPLOADING'
    UPLOAD_COMPLETE = 'UPLOAD_COMPLETE'
    UPLOADING_FAILURE = 'UPLOADING_FAILURE'
    FLASHING = 'FLASHING'
    FLASHING_COMPLETE = 'FLASHING_COMPLETE'
    FLASHING_FAILURE = 'FLASHING_FAILURE'


class Command(str, Enum):
    IDLE = 'idle'
    FIRMWARE_UPLOAD = 'firmware_upload'
    FIRMWARE_FLASH = 'firmware_flash'
    FIRMWARE_CANCEL = 'firmware_cancel'
    INCLUDE = 'include'
    EXCLUDE = 'exclude'
    CONNECTION_CHECK = 'connection_check'


class OwnerEnum(str, Enum):
    UNASSIGNED = 'unassigned'


class Deletion(str, Enum):
    PENDING = 'pending'
    FAILED = 'failed'


class WappstoVersion(str, Enum):
    V2_0 = "2.0"
    V2_1 = "2.1"


class PermissionType(str, Enum):
    """All possible Value Permission Types."""
    READ = 'r'
    WRITE = 'w'
    READWRITE = 'rw'
    WRITEREAD = 'wr'
    NONE = 'none'


class EventStatus(str, Enum):
    OK = 'ok'
    UPDATE = 'update'
    PENDING = 'pending'


class StateType(str, Enum):
    REPORT = 'Report'
    CONTROL = 'Control'


class StateStatus(str, Enum):
    SEND = 'Send'
    PENDING = 'Pending'
    FAILED = 'Failed'


class Level(str, Enum):
    IMPORTANT = 'important'
    ERROR = 'error'
    WARNING = 'warning'
    SUCCESS = 'success'
    INFO = 'info'
    DEBUG = 'debug'


class StatusType(str, Enum):
    PUBLIC_KEY = 'public key'
    MEMORY_INFORMATION = 'memory information'
    DEVICE_DESCRIPTION = 'device description'
    VALUE_DESCRIPTION = 'value description'
    VALUE = 'value'
    PARTNER_INFORMATION = 'partner information'
    ACTION = 'action'
    CALCULATION = 'calculation'
    TIMER = 'timer'
    CALENDAR = 'calendar'
    STATEMACHINE = 'statemachine'
    FIRMWARE_UPDATE = 'firmware update'
    CONFIGURATION = 'configuration'
    EXI = 'exi'
    SYSTEM = 'system'
    APPLICATION = 'application'
    GATEWAY = 'gateway'


class WappstoMetaType(str, Enum):
    """
    All possible Wappsto Meta Types.

    They have a parent child relation in order of:
    network->device->value->state

    Where a 'Network' only contains 'Device's,
    'Device's only contains 'value's, and
    'Value's only contains 'State's.
    """
    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"
    CREATOR = "creator"
    IDLIST = "idlist"
    DELETELIST = "deletelist"


class Connection(BaseModel):
    timestamp: Optional[datetime] = None
    online: Optional[bool] = None

    class Config:
        json_encoders = {
            datetime: timestamp_converter
        }


class WarningItem(BaseModel):
    message: Optional[Optional[str]] = None
    data: Optional[Optional[Dict[str, Any]]] = None
    code: Optional[Optional[int]] = None


class Geo(BaseModel):
    latitude: Optional[str] = None
    longitude: Optional[str] = None
    display_name: Optional[str] = None
    address: Optional[Dict[str, Any]] = None


class BaseMeta(BaseModel):  # Base Meta
    id: Optional[UUID4] = None
    # NOTE: Set in the children-class.
    # #  type: Optional[WappstoMetaType] = None
    version: Optional[WappstoVersion] = None

    manufacturer: Optional[UUID4] = None
    owner: Optional[Union[UUID4, OwnerEnum]] = None
    parent: Optional[UUID4] = None

    created: Optional[datetime] = None
    updated: Optional[datetime] = None
    changed: Optional[datetime] = None

    application: Optional[UUID4] = None
    deletion: Optional[Deletion] = None
    deprecated: Optional[Optional[bool]] = None

    iot: Optional[Optional[bool]] = None
    revision: Optional[Optional[int]] = None
    size: Optional[Optional[int]] = None
    path: Optional[Optional[str]] = None

    oem: Optional[Optional[str]] = None
    accept_manufacturer_as_owner: Optional[Optional[bool]] = None
    redirect: Optional[  # type: ignore
        constr(
            regex=r'^[0-9a-zA-Z_-]+$',  # noqa: F722
            min_length=1,
            max_length=200
        )
    ] = None

    error: Optional[UUID4] = None
    warning: Optional[List[WarningItem]] = None
    trace: Optional[Optional[str]] = None

    set: Optional[List[UUID4]] = None
    contract: Optional[List[UUID4]] = None

    historical: Optional[bool] = None

    class Config:
        json_encoders = {
            datetime: timestamp_converter
        }


class EventlogMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATUS = "eventlog"
    type: Optional[WappstoMetaType] = None

    icon: Optional[Optional[str]] = None
    alert: Optional[List[UUID4]] = None


class StatusMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATUS = "status"
    type: Optional[WappstoMetaType] = None

    icon: Optional[Optional[str]] = None
    alert: Optional[List[UUID4]] = None


class ValueMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        VALUE = "value"
    type: Optional[WappstoMetaType] = None


class StateMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATE = "state"
    type: Optional[WappstoMetaType] = None


class DeviceMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        DEVICE = "device"
    type: Optional[WappstoMetaType] = None

    geo: Optional[Geo] = None


class NetworkMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        NETWORK = "network"
    type: Optional[WappstoMetaType] = None

    geo: Optional[Geo] = None
    connection: Optional[Connection] = None
    accept_test_mode: Optional[bool] = None
    verify_product: Optional[str] = None
    product: Optional[str] = None


class Status(BaseModel):
    message: str
    timestamp: datetime
    data: Optional[str] = None
    level: Level
    type: Optional[StatusType]
    meta: Optional[StatusMeta] = Field(None, title='meta-2.0:create')

    class Config:
        json_encoders = {
            datetime: timestamp_converter
        }


class Info(BaseModel):
    enabled: Optional[bool] = None


class State(BaseModel):
    data: str
    type: Optional[StateType]

    meta: Optional[StateMeta] = Field(None, title='meta-2.0:create')
    status: Optional[StateStatus] = None
    status_payment: Optional[str] = None
    timestamp: Optional[str] = None

    class Config:
        extra = Extra.forbid
        json_encoders = {
            datetime: timestamp_converter,
        }


class EventlogItem(BaseModel):
    message: str
    timestamp: Optional[datetime] = None
    info: Optional[Dict[str, Any]] = None
    level: Level
    type: Optional[str] = None
    meta: Optional[EventlogMeta] = Field(None, title='meta-2.0:create')


class BaseValue(BaseModel):
    name: Optional[str] = None
    type: Optional[str] = None
    description: Optional[str] = None
    period: Optional[str] = None
    delta: Optional[str] = None
    permission: Optional[PermissionType] = None
    status: Optional[EventStatus] = None
    meta: Optional[ValueMeta] = Field(None, title='meta-2.0:create')
    state: Optional[List[Union[State, UUID4]]] = None
    eventlog: Optional[List[Union['EventlogItem', UUID4]]] = None
    info: Optional[Info] = None


class Number(BaseModel):
    min: float
    max: float
    step: float
    mapping: Optional[Dict[str, Any]] = None
    meaningful_zero: Optional[bool] = None
    ordered_mapping: Optional[bool] = None
    si_conversion: Optional[str] = None
    unit: Optional[str] = None


class String(BaseModel):
    max: Optional[conint(ge=1, multiple_of=1)] = None  # type: ignore
    encoding: Optional[str] = None


class Blob(BaseModel):
    max: Optional[conint(ge=1, multiple_of=1)] = None  # type: ignore
    encoding: Optional[str] = None


class Xml(BaseModel):
    xsd: Optional[str] = None
    namespace: Optional[str] = None


class StringValue(BaseValue):
    string: Optional[String] = None

    @root_validator(pre=True)
    def value_type_check(cls, values):
        keys = ["number", "blob", "xml"]
        if any(key in values for key in keys):
            raise TypeError(f"A string value can not contain: {' '.join(keys)}")
        return values


class NumberValue(BaseValue):
    number: Optional[Number] = None

    @root_validator(pre=True)
    def value_type_check(cls, values):
        keys = ["string", "blob", "xml"]
        if any(key in values for key in keys):
            raise TypeError(f"A number value can not contain: {' '.join(keys)}")
        return values


class BlobValue(BaseValue):
    blob: Optional[Blob] = None

    @root_validator(pre=True)
    def value_type_check(cls, values):
        keys = ["number", "string", "xml"]
        if any(key in values for key in keys):
            raise TypeError(f"A blob value can not contain: {' '.join(keys)}")
        return values


class XmlValue(BaseValue):
    xml: Optional[Xml] = None

    @root_validator(pre=True)
    def value_type_check(cls, values):
        keys = ["number", "blob", "blob"]
        if any(key in values for key in keys):
            raise TypeError(f"A xml value can not contain: {' '.join(keys)}")
        return values


class Device(BaseModel):
    name: Optional[str] = None
    control_timeout: Optional[int] = None
    control_when_offline: Optional[bool] = None
    manufacturer: Optional[str] = None
    product: Optional[str] = None
    version: Optional[str] = None
    serial: Optional[str] = None
    description: Optional[str] = None
    protocol: Optional[str] = None
    communication: Optional[str] = None
    included: Optional[str] = None
    inclusion_status: Optional[InclusionStatus] = None
    firmware_status: Optional[FirmwareStatus] = None
    firmware_upload_progress: Optional[str] = None
    firmware_available_version: Optional[str] = None
    command: Optional[Command] = None
    meta: Optional[DeviceMeta] = Field(None, title='meta-2.0:create')
    status: Optional[List[Union[Status, UUID4]]] = None
    value: Optional[
        List[
            Union[
                StringValue,
                NumberValue,
                BlobValue,
                XmlValue,
                UUID4
            ]
        ]
    ] = None
    info: Optional[Info] = None

    # class Config:
    #     extra = Extra.forbid


class Network(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    device: Optional[List[Union[Device, UUID4]]] = None
    meta: Optional[NetworkMeta] = Field(None, title='meta-2.0:create')
    info: Optional[Info] = None

    # class Config:
    #     extra = Extra.forbid


class ApiMetaTypes(str, Enum):
    idlist = "idlist"
    deletelist = "deletelist"


class ApiMetaInfo(BaseModel):
    type: ApiMetaTypes  # Merge with MetaAPIData?
    version: WappstoVersion


class childInfo(BaseModel):
    type: WappstoMetaType
    version: WappstoVersion


class IdList(BaseModel):
    child: List[childInfo]
    id: List[UUID4]
    more: bool
    limit: int
    count: int
    meta: ApiMetaInfo

    class Config:
        extra = Extra.forbid


class DeleteList(BaseModel):
    deleted: List[UUID4]
    code: int
    message: str = "Deleted"
    meta: ApiMetaInfo


WappstoObject = Union[
    Network,
    Device,
    StringValue,
    NumberValue,
    BlobValue,
    XmlValue,
    State,
    IdList,
    DeleteList,
]



from _typeshed import Incomplete

debug: bool
wappstoUrl: Incomplete

def start_session(base_url, username, password): ...
def create_network(session, base_url, product: Incomplete | None = ..., test_mode: bool = ..., reset_manufacturer: bool = ..., dry_run: bool = ...): ...
def claim_network(session, base_url, network_uuid, dry_run: bool = ...): ...
def get_network(session, base_url, network_uuid): ...
def create_certificaties_files(location, creator, dry_run) -> None: ...
def main() -> None: ...



import uuid
import datetime

from enum import Enum
from itertools import zip_longest

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
# from typing import Type
from typing import Union

from pydantic import BaseModel
from pydantic import Extra
from pydantic import parse_obj_as
from pydantic import validator

from .base_schema import BlobValue
from .base_schema import Device
from .base_schema import Network
from .base_schema import NumberValue
from .base_schema import State
from .base_schema import StringValue
from .base_schema import XmlValue
from .base_schema import IdList
from .base_schema import DeleteList


def parwise(values):
    a = iter(values)
    return zip_longest(a, a)


ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

JsonRpc_error_codes = {
    # Rpc Error Code: [HTTP Error Code, "Error String"]
    -32700: [400, "Parse error"],
    -32600: [400, "Invalid Request"],
    -32601: [404, "Method not found"],
    -32602: [400, "Invalid params"],
    -32603: [500, "Internal Server Error"],

    -32000: [404, "Timeout"],
    -32001: [301, "Moved Permanently"],
    -32002: [400, "Bad Request"],
    -32003: [401, "Unauthorized"],
    -32004: [402, "Payment Required"],
    -32005: [403, "Forbidden"],
    -32006: [404, "Not Found"],
    -32007: [405, "Method Not Allowed"],
    -32008: [406, "Not Acceptable"],
    -32009: [408, "Request Timeout"],
    -32010: [409, "Conflict"],
    -32011: [410, "Gone"],
    -32012: [415, "Unsupported Media Type"],
    -32013: [418, "I'm a teapot"],
    -32014: [501, "Not implemented"],
    -32015: [502, "Bad Gateway"],
    -32016: [503, "Service Unavailable"],
    -32017: [504, "Gateway Timeout"],
    -32018: [401, "Queue Limit Reached"]
}


class WappstoObjectType(str, Enum):
    # WappstoMetaType
    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"


ObjectType2BaseModel: Dict[WappstoObjectType, Any] = {
        WappstoObjectType.NETWORK: Network,
        WappstoObjectType.DEVICE: Device,
        WappstoObjectType.VALUE: ValueUnion,
        WappstoObjectType.STATE: State,
}


def url_parser(url: str) -> List[Tuple[WappstoObjectType, Optional[uuid.UUID]]]:
    r_list: List[Tuple[WappstoObjectType, Optional[uuid.UUID]]] = []
    obj_type: Optional[WappstoObjectType] = None
    if url is None:
        raise ValueError("Url need to be Set.")
    parsed_url = url.split("?")[0]
    if parsed_url.startswith("/services/2.0/"):
        parsed_url = parsed_url.replace("/services/2.0", "")
    for selftype, selfid in parwise(parsed_url.split("/")[1:]):
        obj_type = WappstoObjectType(selftype)
        if selfid:
            r_list.append((obj_type, uuid.UUID(selfid)))
        else:
            r_list.append((obj_type, None))
            break
    return r_list


class WappstoMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"


class Success(BaseModel):
    success: bool = True

    class Config:
        extra = Extra.forbid


class Identifier(BaseModel):
    identifier: Optional[str]  # UNSURE: Should this always be there?
    fast: Optional[bool]  # Default: False


class JsonMeta(BaseModel):
    server_send_time: datetime.datetime


class JsonReply(BaseModel):
    value: Optional[Union[
        Device,
        Network,
        State,
        ValueUnion,
        IdList,
        DeleteList,
        bool
    ]]
    meta: JsonMeta

    class Config:
        extra = Extra.forbid

    # @validator("value", pre=True, always=True)
    # def url_data_mapper(
    #     cls, v, values, **kwargs
    # ) -> Optional[Union[
    #         Network,
    #         Device,
    #         ValueUnion,
    #         State,
    #         IdList,
    #         DeleteList,
    #         bool
    #     ]
    # ]:
    #     """Check & enforce the data schema, depended on the method value."""
    #     if v is None or isinstance(v, bool):
    #         return v
    #     if type(v) in ObjectType2BaseModel.values():
    #         return v

    #     # TODO: handle IdList & DeleteList

    #     url_obj = url_parser(values.get('url'))
    #     obj_type = url_obj[-1][0]

    #     model = ObjectType2BaseModel[obj_type]
    #     if model is None:
    #         raise ValueError('Unhandled Object type.')
    #     return parse_obj_as(model, v)


class JsonData(BaseModel):
    url: str
    data: Optional[Any]
    # data: Optional[Union[
    #     Device,
    #     State,
    #     Network,
    #     ValueUnion,
    #     IdList,
    #     DeleteList,
    # ]]
    meta: Optional[Identifier]

    class Config:
        extra = Extra.forbid

    @validator('url')
    def path_check(cls, v, values, **kwargs):
        url_parser(v)
        return v

    @validator("data", pre=True, always=True)
    def url_data_mapper(
        cls, v, values, **kwargs
    ) -> Optional[Union[
            Network,
            Device,
            ValueUnion,
            State,
            IdList,
            DeleteList,
        ]
    ]:
        """Check & enforce the data schema, depended on the method value."""
        if v is None:
            return v
        if type(v) in ObjectType2BaseModel.values():
            return v

        # TODO: handle IdList & DeleteList

        url_obj = url_parser(values.get('url'))
        obj_type = url_obj[-1][0]

        model = ObjectType2BaseModel[obj_type]
        if model is None:
            raise ValueError('Unhandled Object type.')
        return parse_obj_as(model, v)



import abc
from abc import ABC, abstractmethod
from enum import Enum
from threading import Lock
from typing import Any, Callable, Optional, Union

class StatusID(str, Enum):
    CONNECTING: str
    CONNECTED: str
    DISCONNECTING: str
    DISCONNETCED: str

class Connection(ABC, metaclass=abc.ABCMeta):
    send_ready: Lock
    @abstractmethod
    def send(self, data: Union[str, bytes]) -> bool: ...
    @abstractmethod
    def receive(self, parser: Callable[[bytes], Any]) -> Any: ...
    @abstractmethod
    def connect(self) -> Optional[bool]: ...
    @abstractmethod
    def reconnect(self, retry_limit: Optional[int] = ...) -> bool: ...
    @abstractmethod
    def disconnect(self) -> None: ...
    @abstractmethod
    def close(self) -> None: ...



from _typeshed import Incomplete
from datetime import datetime
from enum import Enum
from pydantic import BaseModel, UUID4 as UUID4, conint as conint, constr as constr
from typing import Any, Dict, List, Optional, Union

def timestamp_converter(dt: datetime) -> str: ...

class WappstoMethods(str, Enum):
    DELETE: str
    PUT: str
    POST: str
    GET: str

class InclusionStatus(str, Enum):
    STATUS_DEVICE_INCLUDING: str
    STATUS_DEVICE_INCLUSION_SUCCESS: str
    STATUS_DEVICE_INCLUSION_FAILURE: str
    STATUS_DEVICE_REPORTING: str
    STATUS_DEVICE_REPORT_SUCCESS: str
    STATUS_DEVICE_REPORT_FAILURE: str
    EXCLUDED: str
    INCLUDED: str

class FirmwareStatus(str, Enum):
    UP_TO_DATE: str
    UPDATE_AVAILABLE: str
    UPLOADING: str
    UPLOAD_COMPLETE: str
    UPLOADING_FAILURE: str
    FLASHING: str
    FLASHING_COMPLETE: str
    FLASHING_FAILURE: str

class Command(str, Enum):
    IDLE: str
    FIRMWARE_UPLOAD: str
    FIRMWARE_FLASH: str
    FIRMWARE_CANCEL: str
    INCLUDE: str
    EXCLUDE: str
    CONNECTION_CHECK: str

class OwnerEnum(str, Enum):
    UNASSIGNED: str

class Deletion(str, Enum):
    PENDING: str
    FAILED: str

class WappstoVersion(str, Enum):
    V2_0: str
    V2_1: str

class PermissionType(str, Enum):
    READ: str
    WRITE: str
    READWRITE: str
    WRITEREAD: str
    NONE: str

class EventStatus(str, Enum):
    OK: str
    UPDATE: str
    PENDING: str

class StateType(str, Enum):
    REPORT: str
    CONTROL: str

class StateStatus(str, Enum):
    SEND: str
    PENDING: str
    FAILED: str

class Level(str, Enum):
    IMPORTANT: str
    ERROR: str
    WARNING: str
    SUCCESS: str
    INFO: str
    DEBUG: str

class StatusType(str, Enum):
    PUBLIC_KEY: str
    MEMORY_INFORMATION: str
    DEVICE_DESCRIPTION: str
    VALUE_DESCRIPTION: str
    VALUE: str
    PARTNER_INFORMATION: str
    ACTION: str
    CALCULATION: str
    TIMER: str
    CALENDAR: str
    STATEMACHINE: str
    FIRMWARE_UPDATE: str
    CONFIGURATION: str
    EXI: str
    SYSTEM: str
    APPLICATION: str
    GATEWAY: str

class WappstoMetaType(str, Enum):
    NETWORK: str
    DEVICE: str
    VALUE: str
    STATE: str
    CREATOR: str
    IDLIST: str
    DELETELIST: str

class Connection(BaseModel):
    timestamp: Optional[datetime]
    online: Optional[bool]
    class Config:
        json_encoders: Incomplete

class WarningItem(BaseModel):
    message: Optional[Optional[str]]
    data: Optional[Optional[Dict[str, Any]]]
    code: Optional[Optional[int]]

class Geo(BaseModel):
    latitude: Optional[str]
    longitude: Optional[str]
    display_name: Optional[str]
    address: Optional[Dict[str, Any]]

class BaseMeta(BaseModel):
    id: Optional[UUID4]
    version: Optional[WappstoVersion]
    manufacturer: Optional[UUID4]
    owner: Optional[Union[UUID4, OwnerEnum]]
    parent: Optional[UUID4]
    created: Optional[datetime]
    updated: Optional[datetime]
    changed: Optional[datetime]
    application: Optional[UUID4]
    deletion: Optional[Deletion]
    deprecated: Optional[Optional[bool]]
    iot: Optional[Optional[bool]]
    revision: Optional[Optional[int]]
    size: Optional[Optional[int]]
    path: Optional[Optional[str]]
    oem: Optional[Optional[str]]
    accept_manufacturer_as_owner: Optional[Optional[bool]]
    redirect: Optional[None]
    error: Optional[UUID4]
    warning: Optional[List[WarningItem]]
    trace: Optional[Optional[str]]
    set: Optional[List[UUID4]]
    contract: Optional[List[UUID4]]
    historical: Optional[bool]
    class Config:
        json_encoders: Incomplete

class EventlogMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATUS: str
    type: Optional[WappstoMetaType]
    icon: Optional[Optional[str]]
    alert: Optional[List[UUID4]]

class StatusMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATUS: str
    type: Optional[WappstoMetaType]
    icon: Optional[Optional[str]]
    alert: Optional[List[UUID4]]

class ValueMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        VALUE: str
    type: Optional[WappstoMetaType]

class StateMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        STATE: str
    type: Optional[WappstoMetaType]

class DeviceMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        DEVICE: str
    type: Optional[WappstoMetaType]
    geo: Optional[Geo]

class NetworkMeta(BaseMeta):
    class WappstoMetaType(str, Enum):
        NETWORK: str
    type: Optional[WappstoMetaType]
    geo: Optional[Geo]
    connection: Optional[Connection]
    accept_test_mode: Optional[bool]
    verify_product: Optional[str]
    product: Optional[str]

class Status(BaseModel):
    message: str
    timestamp: datetime
    data: Optional[str]
    level: Level
    type: Optional[StatusType]
    meta: Optional[StatusMeta]
    class Config:
        json_encoders: Incomplete

class Info(BaseModel):
    enabled: Optional[bool]

class State(BaseModel):
    data: str
    type: Optional[StateType]
    meta: Optional[StateMeta]
    status: Optional[StateStatus]
    status_payment: Optional[str]
    timestamp: Optional[str]
    class Config:
        extra: Incomplete
        json_encoders: Incomplete

class EventlogItem(BaseModel):
    message: str
    timestamp: Optional[datetime]
    info: Optional[Dict[str, Any]]
    level: Level
    type: Optional[str]
    meta: Optional[EventlogMeta]

class BaseValue(BaseModel):
    name: Optional[str]
    type: Optional[str]
    description: Optional[str]
    period: Optional[str]
    delta: Optional[str]
    permission: Optional[PermissionType]
    status: Optional[EventStatus]
    meta: Optional[ValueMeta]
    state: Optional[List[Union[State, UUID4]]]
    eventlog: Optional[List[Union['EventlogItem', UUID4]]]
    info: Optional[Info]

class Number(BaseModel):
    min: float
    max: float
    step: float
    mapping: Optional[Dict[str, Any]]
    meaningful_zero: Optional[bool]
    ordered_mapping: Optional[bool]
    si_conversion: Optional[str]
    unit: Optional[str]

class String(BaseModel):
    max: Optional[None]
    encoding: Optional[str]

class Blob(BaseModel):
    max: Optional[None]
    encoding: Optional[str]

class Xml(BaseModel):
    xsd: Optional[str]
    namespace: Optional[str]

class StringValue(BaseValue):
    string: Optional[String]
    def value_type_check(cls, values): ...

class NumberValue(BaseValue):
    number: Optional[Number]
    def value_type_check(cls, values): ...

class BlobValue(BaseValue):
    blob: Optional[Blob]
    def value_type_check(cls, values): ...

class XmlValue(BaseValue):
    xml: Optional[Xml]
    def value_type_check(cls, values): ...

class Device(BaseModel):
    name: Optional[str]
    control_timeout: Optional[int]
    control_when_offline: Optional[bool]
    manufacturer: Optional[str]
    product: Optional[str]
    version: Optional[str]
    serial: Optional[str]
    description: Optional[str]
    protocol: Optional[str]
    communication: Optional[str]
    included: Optional[str]
    inclusion_status: Optional[InclusionStatus]
    firmware_status: Optional[FirmwareStatus]
    firmware_upload_progress: Optional[str]
    firmware_available_version: Optional[str]
    command: Optional[Command]
    meta: Optional[DeviceMeta]
    status: Optional[List[Union[Status, UUID4]]]
    value: Optional[List[Union[StringValue, NumberValue, BlobValue, XmlValue, UUID4]]]
    info: Optional[Info]

class Network(BaseModel):
    name: Optional[str]
    description: Optional[str]
    device: Optional[List[Union[Device, UUID4]]]
    meta: Optional[NetworkMeta]
    info: Optional[Info]

class ApiMetaTypes(str, Enum):
    idlist: str
    deletelist: str

class ApiMetaInfo(BaseModel):
    type: ApiMetaTypes
    version: WappstoVersion

class childInfo(BaseModel):
    type: WappstoMetaType
    version: WappstoVersion

class IdList(BaseModel):
    child: List[childInfo]
    id: List[UUID4]
    more: bool
    limit: int
    count: int
    meta: ApiMetaInfo
    class Config:
        extra: Incomplete

class DeleteList(BaseModel):
    deleted: List[UUID4]
    code: int
    message: str
    meta: ApiMetaInfo
WappstoObject = Union[Network, Device, StringValue, NumberValue, BlobValue, XmlValue, State, IdList, DeleteList]



import abc
from _typeshed import Incomplete
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Union

class OfflineStorage(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def save(self, data: str) -> None: ...
    @abstractmethod
    def load(self, max_count: Optional[int] = ...) -> List[str]: ...

class OfflineStorageFiles(OfflineStorage):
    log: Incomplete
    loc: Incomplete
    suffix: str
    def __init__(self, location: Union[Path, str]) -> None: ...
    def auto_save(self, data: str): ...
    def save(self, data: str) -> None: ...
    def load(self, max_count: Optional[int] = ...) -> List[str]: ...



import uuid
from ..schema.iot_schema import WappstoMethod as WappstoMethod
from ..service.template import ServiceClass as ServiceClass
from ..utils import name_check as name_check
from .device import Device as Device
from _typeshed import Incomplete
from typing import Callable, Optional

class Network:
    schema: Incomplete
    log: Incomplete
    element: Incomplete
    children_uuid_mapping: Incomplete
    children_name_mapping: Incomplete
    cloud_id_mapping: Incomplete
    connection: Incomplete
    def __init__(self, name: str, connection: ServiceClass, network_uuid: uuid.UUID, description: str = ...) -> None: ...
    @property
    def name(self) -> Optional[str]: ...
    @property
    def uuid(self) -> uuid.UUID: ...
    def onChange(self, callback: Callable[[Network], None]) -> Callable[[Network], None]: ...
    def cancelOnChange(self) -> None: ...
    def onCreate(self, callback: Callable[[Network], None]) -> Callable[[Network], None]: ...
    def cancelOnCreate(self) -> None: ...
    def onRefresh(self, callback: Callable[[Network], None]) -> Callable[[Network], None]: ...
    def cancelOnRefresh(self) -> None: ...
    def onDelete(self, callback: Callable[[Network], None]) -> Callable[[Network], None]: ...
    def cancelOnDelete(self) -> None: ...
    def refresh(self) -> None: ...
    def change(self) -> None: ...
    def delete(self) -> None: ...
    def createDevice(self, name: str, manufacturer: Optional[str] = ..., product: Optional[str] = ..., version: Optional[str] = ..., serial: Optional[str] = ..., protocol: Optional[str] = ..., communication: Optional[str] = ..., description: Optional[str] = ...) -> Device: ...
    def close(self) -> None: ...



"""This the the Simple Wappsto Python user-interface to the Wappsto devices."""

# #############################################################################
#                             Modules Import Stuff
# #############################################################################


import __main__
import atexit
import json
import logging
import threading
import uuid

from pathlib import Path
from enum import Enum


from typing import Any, Dict, Optional, Union, Callable


from .modules.network import Network
# from .modules.network import ConnectionStatus
# from .modules.network import ConnectionTypes
# from .modules.network import NetworkChangeType  # NOt needed anymore.
# from .modules.network import NetworkRequestType  # NOt needed anymore.
# from .modules.network import ServiceStatus
# from .modules.network import StatusID

from .modules.device import Device
from .service.template import ServiceClass
from .service.iot_api import IoTAPI

from .modules.value import Value
# from .modules.value import Delta  # Note: Not ready yet!
# from .modules.value import Period  # Note: Not ready yet!
from .modules.value import PermissionType
from .modules.template import ValueTemplate

from .service import template as service

from .connections import protocol as connection

from .utils.offline_storage import OfflineStorage
from .utils.certificateread import certificate_info_extraction
from .utils.offline_storage import OfflineStorageFiles

from .utils import observer
from .utils import name_check

# #############################################################################
#                             __init__ Setup Stuff
# #############################################################################

__version__ = "v0.6.6"
__auther__ = "Seluxit A/S"

__all__ = [
    'Network',
    'Device',
    'Value',
    'onStatusChange',
    'config',
    'createNetwork',
    'connect',
    'disconnect',
    'close',
    'OfflineStorage',
    'service',
    'connection',
    'ValueTemplate',
    'PermissionType'
]


# #############################################################################
#                  Import Stuff for setting up WappstoIoT
# #############################################################################

__log = logging.getLogger("wappstoiot")
__log.addHandler(logging.NullHandler())

# #############################################################################
#                             Status Stuff
# #############################################################################


def onStatusChange(
    StatusID: Union[service.StatusID, connection.StatusID],
    callback: Callable[[Union[service.StatusID, connection.StatusID], Any], None]
):
    """
    Configure an action when the Status have changed.

    def callback(StatusID: StatusID, newStatus: Any):
    """
    observer.subscribe(
        event_name=StatusID,
        callback=callback
    )


# #############################################################################
#                             Config Stuff
# #############################################################################

__config_folder: Path
__the_connection: Optional[ServiceClass] = None
__connection_closed: bool = False
__ping_pong_thread_killed = threading.Event()
__offline_storage_thread_killed = threading.Event()


class ConnectionTypes(str, Enum):
    IOTAPI = "jsonrpc"
    RESTAPI = "HTTPS"


def config(
    config_folder: Union[Path, str] = ".",  # Relative to the main.py-file.
    connection: ConnectionTypes = ConnectionTypes.IOTAPI,
    # JPC_timeout=3
    # mix_max_enforce="warning",  # "ignore", "enforce"
    # step_enforce="warning",  # "ignore", "enforce"
    fast_send: bool = True,  # TODO: jsonrpc.params.meta.fast=true
    # delta_handling="",
    # period_handling="",
    ping_pong_period_sec: Optional[int] = None,  # Period between a RPC ping-pong.
    # # Send: {"jsonrpc":"2.0","method":"HEAD","id":"PING-15","params":{"url":"/services/2.0/network"}}
    # # receive:
    # {"jsonrpc":"2.0","id":"PING-15","result":{"value":true,"meta":{"server_send_time":"2021-12-15T14:33:11.952629Z"}}}
    offline_storage: Union[OfflineStorage, bool] = False,
    # none_blocking=True,  # Whether the post should wait for reply or not.
) -> None:
    """
    Configure the WappstoIoT settings.

    This function call is optional.
    If it is not called, the default settings will be used for WappstoIoT.
    This function will also connect to the WappstoIoT API on call.
    In the cases that this function is not called, the connection will be
    make when an action is make that requests the connection.

    The 'minMaxEnforce' is default set to "Warning" where a warning is
    reading to log, when the value range is outside the minimum & maximum
    range.
    The 'ignore' is where it do nothing when it is outside range.
    The 'enforce' is where the range are enforced to fit the minimum &
    maximum range. Meaning if it is above the maximum it is changed to
    the maximum, if it is below the minimum, it is set to the minimum value.
    """
    global __config_folder
    global __connection_closed
    global __the_connection
    __the_connection = None
    __connection_closed = False

    if not isinstance(config_folder, Path):
        if config_folder == "." and hasattr(__main__, '__file__'):
            __config_folder = Path(__main__.__file__).absolute().parent / Path(config_folder)
        else:
            __config_folder = Path(config_folder)
    else:
        __config_folder = config_folder

    _setup_ping_pong(ping_pong_period_sec)
    _setup_offline_storage(offline_storage)

    if connection == ConnectionTypes.IOTAPI:
        _setup_IoTAPI(__config_folder, fast_send=fast_send)

    # elif connection == ConnectionTypes.RESTAPI:
    #     # TODO: Find & load configs.
    #     configs: Dict[Any, Any] = {}
    #     _setup_RestAPI(__config_folder, configs)  # FIXME:


def _setup_IoTAPI(__config_folder, configs=None, fast_send=False):
    # TODO: Setup the Connection.
    global __the_connection
    kwargs = _certificate_check(__config_folder)
    __the_connection = IoTAPI(**kwargs, fast_send=fast_send)


# def _setup_RestAPI(__config_folder, configs):
#     # TODO: Setup the Connection.
#     global __the_connection
#     token = configs.get("token")
#     login = netrc.netrc().authenticators(configs.end_point)
#     if token:
#         kwargs = {"token": token}
#     elif login:
#         kwargs = {"username": login[0], "password": login[1]}
#     else:
#         raise ValueError("No login was found.")
#     __the_connection = RestAPI(**kwargs, url=configs.end_point)


def _certificate_check(path) -> Dict[str, Path]:
    """Check if the right certificates are at the given path."""
    certi_path = {
        "ca": "ca.crt",
        "crt": "client.crt",
        "key": "client.key",
    }
    r_paths: Dict[str, Path] = {}
    for k, f in certi_path.items():
        r_paths[k] = path / f
        if not r_paths[k].exists():
            raise FileNotFoundError(f"'{f}' was not found in at: {path}")

    return r_paths


def _setup_ping_pong(period_s: Optional[int] = None):
    # TODO: Test me!
    __ping_pong_thread_killed.clear()
    if not period_s:
        return

    # TODO: Need a close check so it do not hold wappstoiot open.
    def _ping():
        __log.debug("Ping-Pong called!")
        nonlocal thread
        global __ping_pong_thread_killed
        if __ping_pong_thread_killed.is_set():
            return
        try:
            thread = threading.Timer(period_s, _ping)
            thread.start()
            __the_connection.ping()
        except Exception:
            __log.exception("Ping-Pong:")
    thread = threading.Timer(period_s, _ping)
    thread.daemon = True
    thread.start()
    atexit.register(lambda: thread.cancel())
    # atexit.register(lambda: __ping_pong_thread_killed.set())


def _setup_offline_storage(
    offlineStorage: Union[OfflineStorage, bool],
) -> None:
    global __the_connection
    global __offline_storage_thread_killed
    __ping_pong_thread_killed.clear()

    if offlineStorage is False:
        return
    # if offlineStorage is True:
    offline_storage: OfflineStorage = OfflineStorageFiles(
        location=__config_folder
    ) if offlineStorage is True else offlineStorage
    # else:
    #     offline_storage: OfflineStorage = offlineStorage

    observer.subscribe(
        service.StatusID.SENDERROR,
        lambda _, data: offline_storage.save(data.json(exclude_none=True)) if data else None
    )

    def _resend_logic(status, status_data):
        nonlocal offline_storage
        global __offline_storage_thread_killed
        __log.debug(f"Resend called with: status={status}")
        try:
            __log.debug("Resending Offline data")
            while not __offline_storage_thread_killed.is_set():
                data = offline_storage.load(10)
                if not data:
                    return

                s_data = [json.loads(d) for d in data]
                __log.debug(f"Sending Data: {s_data}")
                __the_connection._resend_data(
                    json.dumps(s_data)
                )

        except Exception:
            __log.exception("Resend Logic")

    observer.subscribe(
        connection.StatusID.CONNECTED,
        _resend_logic
    )


# #############################################################################
#                             Create Stuff
# #############################################################################

def createNetwork(
    name: str,
    description: str = "",
) -> Network:
    """
    Create a new Wappsto Network.

    A Wappsto Network is references to the main grouping, of which multiple
    device are connected.
    """
    global __config_folder
    global __the_connection

    illegal_chars: str = name_check.illegal_characters(name)

    if illegal_chars:
        raise ValueError(
            f"Given name contain a illegal character: {illegal_chars}"
            f"May only contain: {name_check.wappsto_letters}"
        )

    if not __the_connection:
        config()

    if not __config_folder:
        __config_folder = Path('.')

    cer = certificate_info_extraction(crt_path=__config_folder / "client.crt")
    network_uuid = uuid.UUID(cer.get('subject', {}).get('commonName'))

    atexit.register(close)

    return Network(
        name=name,
        connection=__the_connection,
        network_uuid=network_uuid,
        description=description
    )

# -------------------------------------------------------------------------
#   Connection methods
# -------------------------------------------------------------------------


def connect():
    """NOT Implemented yet."""
    pass


def disconnect():
    """NOT Implemented yet."""
    pass


def close():
    """Close down the connection to wappsto."""
    atexit.unregister(close)
    __ping_pong_thread_killed.set()
    __offline_storage_thread_killed.set()
    # atexit._run_exitfuncs()
    global __connection_closed
    global __the_connection

    if not __connection_closed and __the_connection is not None:
        __log.info("Closing Wappsto IoT")
        __the_connection.close()
        __connection_closed = True
    # Disconnect
    pass



import datetime
import uuid
from .base_schema import BlobValue as BlobValue, DeleteList as DeleteList, Device as Device, IdList as IdList, Network as Network, NumberValue as NumberValue, State as State, StringValue as StringValue, XmlValue as XmlValue
from _typeshed import Incomplete
from enum import Enum
from pydantic import BaseModel
from typing import Any, Dict, List, Optional, Tuple, Union

def parwise(values): ...
ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]
JsonRpc_error_codes: Incomplete

class WappstoObjectType(str, Enum):
    NETWORK: str
    DEVICE: str
    VALUE: str
    STATE: str

ObjectType2BaseModel: Dict[WappstoObjectType, Any]

def url_parser(url: str) -> List[Tuple[WappstoObjectType, Optional[uuid.UUID]]]: ...

class WappstoMethod(str, Enum):
    GET: str
    POST: str
    PATCH: str
    PUT: str
    DELETE: str
    HEAD: str

class Success(BaseModel):
    success: bool
    class Config:
        extra: Incomplete

class Identifier(BaseModel):
    identifier: Optional[str]
    fast: Optional[bool]

class JsonMeta(BaseModel):
    server_send_time: datetime.datetime

class JsonReply(BaseModel):
    value: Optional[Union[Device, Network, State, ValueUnion, IdList, DeleteList, bool]]
    meta: JsonMeta
    class Config:
        extra: Incomplete

class JsonData(BaseModel):
    url: str
    data: Optional[Any]
    meta: Optional[Identifier]
    class Config:
        extra: Incomplete
    def path_check(cls, v, values, **kwargs): ...
    def url_data_mapper(cls, v, values, **kwargs) -> Optional[Union[Network, Device, ValueUnion, State, IdList, DeleteList]]: ...



from uuid import UUID
from enum import Enum

from typing import Callable
from typing import Union
from typing import Optional

from abc import ABC
from abc import abstractmethod

from ..schema.base_schema import BlobValue
from ..schema.base_schema import Device
from ..schema.base_schema import Network
from ..schema.base_schema import NumberValue
from ..schema.base_schema import State
from ..schema.base_schema import StringValue
from ..schema.base_schema import XmlValue
from ..schema.base_schema import WappstoMethods


class StatusID(str, Enum):
    IDLE = "idle"
    # BATCHING = "batching"
    SENDING = "sending"
    SENDERROR = "send Error"
    SEND = "send"
    ERROR = "Error Reply"


class ServiceClass(ABC):

    # #########################################################################
    #                               Network API
    # #########################################################################

    @abstractmethod
    def subscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethods], None]
    ):
        pass

    @abstractmethod
    def post_network(self, data) -> bool:
        # url=f"/services/2.0/network/{uuid}",
        pass

    @abstractmethod
    def put_network(self, uuid: UUID, data) -> bool:
        pass

    @abstractmethod
    def get_network(self, uuid: UUID) -> Union[Network, None]:
        pass

    @abstractmethod
    def delete_network(self, uuid: UUID) -> bool:
        pass

    # #########################################################################
    #                                Device API
    # #########################################################################

    @abstractmethod
    def subscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethods], None]
    ):
        pass

    @abstractmethod
    def post_device(self, network_uuid: UUID, data: Device) -> bool:
        # url=f"/services/2.0/{uuid}/device",
        pass

    @abstractmethod
    def put_device(self, uuid: UUID, data: Device) -> bool:
        pass

    @abstractmethod
    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        pass

    @abstractmethod
    def get_device(self, uuid: UUID) -> Union[Device, None]:
        pass

    @abstractmethod
    def delete_device(self, uuid: UUID) -> bool:
        pass

    # #########################################################################
    #                                 Value API
    # #########################################################################

    ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

    @abstractmethod
    def subscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethods], None]
    ):
        pass

    @abstractmethod
    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool:
        # url=f"/services/2.0/{uuid}/value",
        pass

    @abstractmethod
    def put_value(self, uuid: UUID, data: ValueUnion) -> bool:
        pass

    @abstractmethod
    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        pass

    @abstractmethod
    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]:
        pass

    @abstractmethod
    def delete_value(self, uuid: UUID) -> bool:
        pass

    # #########################################################################
    #                                State API
    # #########################################################################

    @abstractmethod
    def subscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethods], None]
    ):
        pass

    @abstractmethod
    def post_state(self, value_uuid: UUID, data: State) -> bool:
        # url=f"/services/2.0/{uuid}/state",
        pass

    @abstractmethod
    def put_state(self, uuid: UUID, data: State) -> bool:
        pass

    @abstractmethod
    def get_state(self, uuid: UUID) -> Union[State, None]:
        pass

    @abstractmethod
    def delete_state(self, uuid: UUID) -> bool:
        pass



import uuid
from ..schema.base_schema import PermissionType as PermissionType
from ..schema.iot_schema import WappstoMethod as WappstoMethod
from ..service.template import ServiceClass as ServiceClass
from ..utils import name_check as name_check
from .network import Network as Network
from .template import ValueBaseType as ValueBaseType, ValueTemplate as ValueTemplate, valueSettings as valueSettings
from .value import Value as Value
from _typeshed import Incomplete
from enum import Enum
from typing import Callable, Dict, Optional, Union

class RequestType(str, Enum):
    refresh: str
    delete: str

class ChangeType(str, Enum):
    value: str
    name: str
    manufacturer: str
    product: str
    version: str
    serial: str
    description: str

class Device:
    schema: Incomplete
    log: Incomplete
    parent: Incomplete
    element: Incomplete
    children_uuid_mapping: Incomplete
    children_name_mapping: Incomplete
    cloud_id_mapping: Incomplete
    connection: Incomplete
    def __init__(self, parent: Network, device_uuid: Optional[uuid.UUID], name: Optional[str] = ..., manufacturer: Optional[str] = ..., product: Optional[str] = ..., version: Optional[str] = ..., protocol: Optional[str] = ..., communication: Optional[str] = ..., serial: Optional[str] = ..., description: Optional[str] = ...) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def uuid(self) -> uuid.UUID: ...
    def onDelete(self, callback: Callable[[Device], None]) -> Callable[[Device], None]: ...
    def cancelOnDelete(self) -> None: ...
    def onRefresh(self, callback: Callable[[Device], None]) -> Callable[[Device], None]: ...
    def cancelOnRefresh(self) -> None: ...
    def onChange(self, callback: Callable[[Device], None]) -> Callable[[Device], None]: ...
    def cancelOnChange(self) -> None: ...
    def onCreate(self, callback: Callable[[Device], None]) -> Callable[[Device], None]: ...
    def cancelOnCreate(self) -> None: ...
    def refresh(self) -> None: ...
    def change(self, change_type: ChangeType) -> None: ...
    def delete(self) -> None: ...
    def createNumberValue(self, name: str, *, permission: PermissionType, type: str, min: Union[int, float], max: Union[int, float], step: Union[int, float], unit: str, description: Optional[str] = ..., si_conversion: Optional[str] = ..., period: Optional[int] = ..., delta: Optional[Union[int, float]] = ..., mapping: Optional[Dict[str, str]] = ..., meaningful_zero: Optional[bool] = ..., ordered_mapping: Optional[bool] = ...) -> Value: ...
    def createStringValue(self, name: str, *, permission: PermissionType, type: str, max: Union[int, float], encoding: Optional[str] = ..., description: Optional[str] = ..., period: Optional[int] = ..., delta: Optional[Union[int, float]] = ...) -> Value: ...
    def createBlobValue(self, name: str, *, permission: PermissionType, type: str, max: Union[int, float], encoding: Optional[str] = ..., description: Optional[str] = ..., period: Optional[int] = ..., delta: Optional[Union[int, float]] = ...) -> Value: ...
    def createXmlValue(self, name: str, *, permission: PermissionType, type: str, xsd: Optional[str] = ..., namespace: Optional[str] = ..., description: Optional[str] = ..., period: Optional[int] = ..., delta: Optional[Union[int, float]] = ...) -> Value: ...
    def createValue(self, name: str, permission: PermissionType, value_template: ValueTemplate, description: Optional[str] = ...) -> Value: ...
    def close(self) -> None: ...



import copy
import json
import logging
import re
import threading

from uuid import UUID
from pathlib import Path

from concurrent.futures import ThreadPoolExecutor

from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import slxjsonrpc
from slxjsonrpc.schema.jsonrpc import ErrorModel

from .template import StatusID
from .template import ServiceClass

from ..schema.base_schema import BlobValue
from ..schema.base_schema import Device
from ..schema.base_schema import Network
from ..schema.base_schema import NumberValue
from ..schema.base_schema import State
from ..schema.base_schema import StringValue
from ..schema.base_schema import XmlValue
from ..schema.base_schema import WappstoObject
from ..schema.base_schema import IdList

from ..schema.iot_schema import JsonData
from ..schema.iot_schema import Identifier
from ..schema.iot_schema import JsonReply
from ..schema.iot_schema import Success
from ..schema.iot_schema import WappstoMethod
# from ..schema.iot_scgema import WappstoObjectType

from ..utils.certificateread import certificate_info_extraction
from ..utils import observer
from ..utils import tracer

from ..connections.sslsocket import TlsSocket
from ..connections.protocol import Connection


ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]


# POST   -> onCreate
# GET    -> onRefresh
# PUT    -> onChange
# DELETE -> onDelete

# create  -> POST
# refresh -> GET
# change  -> PUT
# delete  -> DELETE


class IoTAPI(ServiceClass):

    wappstoPort = {
        "dev": 52005,
        "qa": 53005,
        "staging": 54005,
        "prod": 443
    }

    def __init__(
        self,
        ca: Path,
        crt: Path,
        key: Path,
        worker_count: int = 2,
        fast_send: bool = False,
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())
        self.ca = ca
        self.crt = crt
        self.key = key

        self.timeout = 3

        self.fast_send = fast_send

        self.addr, self.port = self._url_gen(self.crt)

        self.connection: Connection

        self.connection = TlsSocket(
            address=self.addr,
            port=self.port,
            ca=self.ca,
            crt=self.crt,
            key=self.key
        )

        params = {
            x: Union[Success, JsonData] for x in WappstoMethod
        }

        result = {
            WappstoMethod.GET: JsonReply,
            WappstoMethod.POST: JsonReply,
            WappstoMethod.PUT: JsonReply,
            WappstoMethod.DELETE: JsonReply,
            # WappstoMethod.PATCH: JsonReply,
            WappstoMethod.HEAD: JsonReply,
        }

        self.subscribers: Dict[
            UUID,
            List[Callable[[WappstoObject, WappstoMethod], None]]
        ] = {}

        method_cb = {
            WappstoMethod.GET: self._get,
            WappstoMethod.POST: self._post,
            # WappstoMethod.PATCH: self._patch,
            WappstoMethod.PUT: self._put,
            WappstoMethod.DELETE: self._delete,
            # WappstoMethod.HEAD: self._head,
        }

        self.connection.connect()

        self.jsonrpc = slxjsonrpc.SlxJsonRpc(
            methods=WappstoMethod,
            method_cb=method_cb,
            result=result,
            params=params,
        )

        self.killed = threading.Event()
        self.workers = ThreadPoolExecutor(max_workers=worker_count)

        self.workers.submit(self._receive_handler)

    def close(self):
        self.killed.set()
        self.log.debug("Closing Connection.")
        self.connection.close()
        self.log.debug("Closing Workers")
        self.workers.shutdown()
        self.log.debug("IoTAPI Closed.")

    # #########################################################################
    #                              Helper Methods
    # #########################################################################

    def _url_gen(self, crt):
        cer = certificate_info_extraction(crt_path=crt)
        endpoint = cer.get('issuer', {}).get('commonName')
        port = self.wappstoPort.get(endpoint.split('.')[0], 443)
        if endpoint.split('.')[0] in self.wappstoPort.keys():
            addr = endpoint
        else:
            addr = f"collector.{endpoint}"

        if re.search(r'^[A-Za-z0-9+.-]+://', addr):
            addr = addr.split("://", maxsplit=1)[-1]
        # if not re.search(r':[0-9]+', addr):
        #     addr = f'{addr}:{port}'
        return (addr, port)

    def _receive_handler(self):
        self.log.debug("Receive Handler Started!")
        while not self.killed.is_set():
            try:
                data = self.connection.receive(parser=json.loads)

                if not data:
                    continue

                if not isinstance(data, list):
                    data = [data]
                for elemt in data:
                    _ids = [elemt.get('id', elemt) for elemt in data]
                    self.log.debug(f"Package received: {elemt.get('id', elemt)}")

                trace = tracer.Trace.trace_list_check(
                    jsonrpc_elemts=data,
                    name="Wappsto IoT Receive Thread"
                )

                reply = self.jsonrpc.parser(data)
                self.log.debug(f"Package ID: {_ids}; Reply: {reply}")

                if not reply:
                    continue

                self._send_logic(reply)
                # UNSURE: How do we check if it was send?
                observer.post(StatusID.SEND, reply)

                self._trace_send_logic(trace, reply)

            except Exception:
                self.log.exception("Receive Handler Error:")
        self.log.debug("Receive Handler Stopped!")

    def _trace_send_logic(self, trace_list, reply_list) -> None:
        if not trace_list:
            return
        for t, d in zip(trace_list, reply_list.__root__):
            if not isinstance(d, ErrorModel):
                t.send_ok(name="Wappsto IoT Receive Thread")
            else:
                t.send_failed(name="Wappsto IoT Receive Thread")

    def _send_logic(self, data, _id=None):
        # NOTE (MBK): Something do not work here!
        # if not data:
        #     return
        with self.connection.send_ready:  # NOTE: Waiting here, until ready!
            with self.jsonrpc.batch():
                batch_size = self.jsonrpc.batch_size()
                if batch_size:
                    send_data = self.jsonrpc.get_batch_data(data)
                    self.log.debug(f"Batching: {batch_size}")
                else:
                    send_data = copy.copy(data)

                if isinstance(send_data, slxjsonrpc.RpcBatch):
                    _id = "; ".join(x.id for x in send_data.__root__)
                elif send_data:
                    _id = send_data.id

                if _id:
                    self.log.debug(f"Package ID: {_id};")

                if send_data:
                    observer.post(StatusID.SENDING, send_data)
                    self.connection.send(send_data.json(exclude_none=True))

    def _resend_data(self, data):
        j_data = json.loads(data)
        _cb_event = threading.Event()

        if not isinstance(j_data, list):
            j_data = [j_data]

        for l_data in j_data:
            rpc_id = l_data.get('id')

            if l_data.get('params'):
                s_data = self.jsonrpc.create_request(
                    callback=lambda data: _cb_event.set(),
                    error_callback=lambda err_data: _cb_event.set(),
                    method=l_data.get('method'),
                    params=l_data.get('params')
                )
                self._send_logic(s_data)
            elif l_data.get('result'):
                self.jsonrpc._id_cb[rpc_id] = lambda data: _cb_event.set()
                self.jsonrpc._id_error_cb[rpc_id] = lambda err_data: _cb_event.set()
                self.jsonrpc._id_method[rpc_id] = l_data.get('method')
                # self.jsonrpc._add_result_handling(  # NOTE: slxJsonRpc v0.8.1.
                #     method=l_data.get('method'),
                #     _id=rpc_id,
                #     callback=lambda data: _cb_event(),
                #     error_callback=lambda err_data: _cb_event(),
                # )
                s_data = self.jsonrpc._parse_data(l_data)
                self._send_logic(s_data)

            self.log.debug(f"--CALLBACK Ready! {rpc_id}")
            if _cb_event.wait(timeout=self.timeout):
                self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
                observer.post(StatusID.SEND, s_data)
            else:
                self.log.debug(f"--CALLBACK None! {rpc_id}")
                observer.post(StatusID.SENDERROR, s_data)
        return None

    # -------------------------------------------------------------------------
    #                               API Helpers
    # -------------------------------------------------------------------------

    def _no_reply_send(self, data, url, method) -> bool:
        j_data = JsonData(
            data=data,
            url=url,
            meta=Identifier(fast=True) if self.fast_send and method != WappstoMethod.GET else None
        )

        self.log.debug(f"Sending for: {url}")

        _cb_event = threading.Event()
        _err_data: Optional[ErrorModel] = None

        def _err_callback(err_data: ErrorModel):
            nonlocal _err_data
            nonlocal _cb_event
            _err_data = err_data
            _cb_event.set()

        rpc_data = self.jsonrpc.create_request(
            method=method,
            callback=lambda data: _cb_event.set(),
            error_callback=_err_callback,
            params=j_data
        )

        rpc_id = getattr(rpc_data, 'id', None)

        self._send_logic(rpc_data)

        self.log.debug(f"--CALLBACK Ready! {rpc_id}")
        if _cb_event.wait(timeout=self.timeout):
            if _err_data:
                self.log.debug(f"--CALLBACK Error! {_err_data}")
                observer.post(StatusID.ERROR, _err_data)
                # raise ConnectionError(_err_data.message)
                return False
            self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
            observer.post(StatusID.SEND, rpc_data)
            return True
        observer.post(StatusID.SENDERROR, rpc_data)
        self.log.debug(f"--CALLBACK None! {rpc_id}")
        return False

    def _reply_send(
        self,
        data,
        url,
        method
    ) -> Optional[Union[Device, Network, ValueUnion, State, IdList]]:
        j_data = JsonData(
            data=data,
            url=url,
            meta=Identifier(fast=True) if self.fast_send and method != WappstoMethod.GET else None
        )

        self.log.debug(f"Sending for: {url}")

        _cb_event = threading.Event()
        _err_data: ErrorModel = None

        def _err_callback(err_data: ErrorModel):
            nonlocal _err_data
            nonlocal _cb_event
            _err_data = err_data
            _cb_event.set()

        _data: Optional[JsonReply] = None

        def _data_callback(data: JsonReply):
            nonlocal _data
            nonlocal _cb_event
            _data = data
            _cb_event.set()

        rpc_data = self.jsonrpc.create_request(
            method=method,
            callback=_data_callback,
            error_callback=_err_callback,
            params=j_data
        )

        rpc_id = getattr(rpc_data, 'id', None)

        self._send_logic(rpc_data)

        self.log.debug(f"--CALLBACK Ready! {rpc_id}")
        if _cb_event.wait(timeout=self.timeout):
            if _data:
                self.log.debug(f"--CALLBACK EVENT! {rpc_id}")
                observer.post(StatusID.SEND, rpc_data)
                return _data.value
            if _err_data:
                self.log.warning(f"--CALLBACK Error! {_err_data}")
                observer.post(StatusID.ERROR, _err_data)
                # raise ConnectionError(_err_data.message)
                return None
        self.log.debug(f"--CALLBACK None! {rpc_id}")
        observer.post(StatusID.SENDERROR, rpc_data)
        return None

    # -------------------------------------------------------------------------
    #                              Callback Helpers
    # -------------------------------------------------------------------------

    def _cb_handler(self, data: JsonData, method: WappstoMethod):
        object_uuid = UUID(data.url.split("/")[-1])
        self.log.debug(f"Object UUID: {object_uuid}")
        for cb in self.subscribers.get(object_uuid, [self._default_cb]):
            self.workers.submit(cb, data.data, method)
            self.log.debug(f"Submitted to Worker: {cb}")

    def _default_cb(self, data: WappstoObject, method: WappstoMethod):
        self.log.warning(
            f"No callback found for method: {method}; data {data}"
        )

    # #########################################################################
    #                              Callback Methods
    # #########################################################################

    def _get(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_get: Called!")
        self._cb_handler(data=data, method=WappstoMethod.GET)
        return Success()

    def _post(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_post: Called!")
        self._cb_handler(data=data, method=WappstoMethod.POST)
        return Success()

    def _put(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_put: Called!")
        self._cb_handler(data=data, method=WappstoMethod.PUT)
        return Success()

    def _delete(self, data: JsonData) -> Union[Success, JsonData]:
        # Add data and callback-function to a threadpool.
        self.log.debug("_delete: Called!")
        self._cb_handler(data=data, method=WappstoMethod.DELETE)
        return Success()

    # def _patch(self, data: JsonData) -> Union[Success, JsonData]:
    #     # Add data and callback-function to a threadpool.
    #     self.log.debug("_patch: Called!")
    #     self._cb_handler(data=data, method=WappstoMethod.PATCH)
    #     return Success()

    # def _head(self, data: JsonData) -> Union[Success, JsonData]:
    #     # Add data and callback-function to a threadpool.
    #     self.log.debug("_head: Called!")
    #     self._cb_handler(data=data, method=WappstoMethod.HEAD)
    #     return Success()

    # #########################################################################
    #                               Helper API
    # #########################################################################

    def ping(self):
        return self._no_reply_send(
            data=None,
            url="/network",
            method=WappstoMethod.HEAD
        )

    # #########################################################################
    #                               Network API
    # #########################################################################

    def subscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ):
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethod], None]
    ):
        self.subscribers.get(uuid, []).remove(callback)

    def post_network(self, data: Network) -> bool:
        # url=f"/services/2.0/network",
        return self._no_reply_send(
            data=data,
            url="/network/",
            method=WappstoMethod.POST
        )

    def put_network(self, uuid: UUID, data: Network) -> bool:
        # url=f"/services/2.0/network/{uuid}",
        return self._no_reply_send(
            data=data,
            url=f"/network/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_network(self, uuid: UUID) -> Optional[Network]:
        # url=f"/services/2.0/network/{uuid}",
        return self._reply_send(
            data=None,  # NOTE: Should be nonexistent or Null.
            url=f"/network/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_network(self, uuid: UUID) -> bool:
        # url=f"/services/2.0/network/{uuid}",
        return self._no_reply_send(
            data=None,  # NOTE: Should be nonexistent or Null.
            url=f"/network/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                Device API
    # #########################################################################

    def subscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ):
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ):
        self.subscribers.get(uuid, []).remove(callback)

    def post_device(self, network_uuid: UUID, data: Device) -> bool:
        # url=f"/services/2.0/{uuid}/device",
        return self._no_reply_send(
            data=data,
            url=f"/network/{network_uuid}/device/",
            method=WappstoMethod.POST
        )

    def put_device(self, uuid: UUID, data: Device) -> bool:
        # url=f"/services/2.0/device/{uuid}",
        return self._no_reply_send(
            data=data,
            url=f"/device/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        # /network/{uuid}/device?this_name==X
        key, value = list(kwargs.items())[0]
        url = f"/network/{network_uuid}/device?this_{key}=={value}"
        data: IdList = self._reply_send(
            data=None,
            url=url,
            method=WappstoMethod.GET
        )

        temp = getattr(data, "id", None)
        if not temp:
            return None
        return temp[0]

    def get_device(self, uuid: UUID) -> Union[Device, None]:
        # url=f"/services/2.0/device/{uuid}",
        return self._reply_send(
            data=None,
            url=f"/device/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_device(self, uuid: UUID) -> bool:
        # url=f"/services/2.0/device/{uuid}",
        return self._no_reply_send(
            data=None,
            url=f"/device/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                 Value API
    # #########################################################################

    def subscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethod], None]
    ):
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ):
        self.subscribers.get(uuid, []).remove(callback)

    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool:
        # url=f"/services/2.0/{uuid}/value",
        return self._no_reply_send(
            data=data,
            url=f"/device/{device_uuid}/value/",
            method=WappstoMethod.POST
        )

    def put_value(self, uuid: UUID, data: ValueUnion) -> bool:
        # url=f"/services/2.0/value/{uuid}",
        return self._no_reply_send(
            data=data,
            url=f"/value/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]:
        # /network/{uuid}/device?this_name==X
        key, value = list(kwargs.items())[0]
        url = f"/device/{device_uuid}/value?this_{key}=={value}"
        data: IdList = self._reply_send(
            data=None,
            url=url,
            method=WappstoMethod.GET
        )

        temp = getattr(data, "id", None)
        if not temp:
            return None
        return temp[0]

    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]:
        # url=f"/services/2.0/value/{uuid}",
        return self._reply_send(
            data=None,
            url=f"/value/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_value(self, uuid: UUID) -> bool:
        # url=f"/services/2.0/value/{uuid}",
        return self._no_reply_send(
            data=None,
            url=f"/value/{uuid}",
            method=WappstoMethod.DELETE
        )

    # #########################################################################
    #                                State API
    # #########################################################################

    def subscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethod], None]
    ):
        self.subscribers.setdefault(uuid, []).append(callback)

    def unsubscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethod], None]
    ):
        self.subscribers.get(uuid, []).remove(callback)

    def post_state(self, value_uuid: UUID, data: State) -> bool:
        # url=f"/services/2.0/{uuid}/state",
        return self._no_reply_send(
            data=data,
            url=f"/value/{value_uuid}/state/",
            method=WappstoMethod.POST
        )

    def put_state(self, uuid: UUID, data: State) -> bool:
        # url=f"/services/2.0/state/{uuid}",
        return self._no_reply_send(
            data=data,
            url=f"/state/{uuid}",
            method=WappstoMethod.PUT
        )

    def get_state(self, uuid: UUID) -> Union[State, None]:
        return self._reply_send(
            data=None,
            url=f"/state/{uuid}",
            method=WappstoMethod.GET
        )

    def delete_state(self, uuid: UUID) -> bool:
        # url=f"/services/2.0/state/{uuid}",
        return self._no_reply_send(
            data=None,
            url=f"/state/{uuid}",
            method=WappstoMethod.DELETE
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A helper script to generate the needed certificates for wappsto.

Certificates are use for the Wappsto IoT library to identify itself on
the wappsto platform, through a secure connection.
"""
import argparse
import getpass
import json
import pathlib
import sys
import uuid

import requests


debug = False


wappstoUrl = {
    "dev": "dev.wappsto.com",
    "qa": "qa.wappsto.com",
    "staging": "staging.wappsto.com",
    "prod": "wappsto.com",
}


def _log_request_error(data):
    if debug:
        print("Sent data     :")
        print(" - URL         : {}".format(data.request.url))
        print(" - Headers     : {}".format(data.request.headers))
        print(" - Request Body: {}".format(
            json.dumps(data.request.body, indent=4, sort_keys=True))
        )

        print("")
        print("")

        print("Received data :")
        print(" - URL         : {}".format(data.url))
        print(" - Headers     : {}".format(data.headers))
        print(" - Status code : {}".format(data.status_code))
        try:
            print(" - Request Body: {}".format(
                json.dumps(json.loads(data.text), indent=4, sort_keys=True))
            )
        except (AttributeError, json.JSONDecodeError):
            pass
    try:
        err = json.loads(data.text)
    except Exception:
        err = data.text
    else:
        err = err.get('message', f"Unknown Error: http error: {data.status_code}")
    print(f"\t{err}")
    exit(-2)


def _start_session(base_url, username, password):
    session_json = {
        "username": username,
        "password": password,
        "remember_me": False
    }

    url = f"https://{base_url}/services/session"
    headers = {"Content-type": "application/json"}
    data = json.dumps(session_json)

    rdata = requests.post(
        url=url,
        headers=headers,
        data=data
    )

    if rdata.status_code >= 300:
        print("\nAn error occurred during login:")
        _log_request_error(rdata)

    rjson = json.loads(rdata.text)

    print(rjson)

    return rjson["meta"]["id"]


def _create_network(
    session,
    base_url,
    # network_uuid=None,
    product=None,
    test_mode=False,
    reset_manufacturer=False,
    dry_run=False
):
    # Should take use of the more general functions.
    request = {}

    url = f"https://{base_url}/services/2.1/creator"
    headers = {
        "Content-type": "application/json",
        "X-session": str(session)
    }

    data = json.dumps(request)

    if not dry_run:
        rdata = requests.post(
            url=url,
            headers=headers,
            data=data
        )

        if rdata.status_code >= 300:
            print("\nAn error occurred during Certificate retrieval:")
            _log_request_error(rdata)

        rjson = json.loads(rdata.text)
    else:
        rjson = {
            'network': {'id': 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'},
            'ca': "NOTHING",
            'certificate': "NOTHING",
            'private_key': "NOTHING",
            'meta': {'id': 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'}
        }
        print("\nDry-run: Fake Certificates created!")

    print(f"\nCertificate generated for new network:\t{rjson['network']['id']}")

    return rjson


def _claim_network(session, base_url, network_uuid, dry_run=False):
    url = f"https://{base_url}/services/2.0/network/{network_uuid}"

    headers = {
        "Content-type": "application/json",
        "X-session": str(session)
    }

    if not dry_run:
        rdata = requests.post(
            url=url,
            headers=headers,
            data="{}"
        )

        if rdata.status_code >= 300:
            print("\nAn error occurred during claiming the network:")
            _log_request_error(rdata)
        rjson = json.loads(rdata.text)
    else:
        rjson = {
            'device': [],
            'meta': {
                'id': network_uuid,
                'type': 'network',
                'version': '2.0'
            }
        }
        print("\nFake Claiming the Network.")

    print(f"\nNetwork: {network_uuid} have been claimed.")
    return rjson


def _get_network(session, base_url, network_uuid):
    url = f"https://{base_url}/services/2.1/creator?this_network.id={network_uuid}"
    headers = {
        "Content-type": "application/json",
        "X-session": str(session)
    }

    rdata = requests.get(
        url=url,
        headers=headers
    )

    if rdata.status_code >= 300:
        print("\nAn error occurred during Certificate retrieval:")
        _log_request_error(rdata)
    data = json.loads(rdata.text)

    if not data['id']:
        if 'message' in data.keys():
            print(f"{data['message']}")
        else:
            print(f"UnKnown Error: {data}")
        exit(-3)
    creator_id = data['id'][0]
    url = f"https://{base_url}/services/2.1/creator/{creator_id}"

    rdata = requests.get(
        url=url,
        headers=headers
    )

    if rdata.status_code >= 300:
        print("\nAn error occurred during Certificate retrieval:")
        _log_request_error(rdata)

    rjson = json.loads(rdata.text)

    print(f"\nCertificate retrieved for network:\t{rjson['network']['id']}")

    return rjson


def _create_certificaties_files(location, creator, dry_run):
    creator["ca"], creator["certificate"], creator["private_key"]

    if not dry_run:
        location.mkdir(exist_ok=True)
        try:
            with open(location / "ca.crt", "w") as file:
                file.write(creator["ca"])
            with open(location / "client.crt", "w") as file:
                file.write(creator["certificate"])
            with open(location / "client.key", "w") as file:
                file.write(creator["private_key"])
        except Exception as err:
            print("\nAn error occurred while saving Certificates:")
            print(f"\t{err}")
            print("\nWhen fixed you can recreate the certificate file with the --recreate option.")
            exit(-3)
    else:
        print("\nDry-run: Fake Save done!")
    print(f"\nLocation of generated certificates:\t{location.absolute()}")


def main():
    """
    Main logic of the program.

    Parses the terminal argument, and execute the thereby given commands.
    """
    global debug

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dry_run",
        action='store_true',
        help="Run the Script, without making the files & folders.",
    )
    parser.add_argument(
        "--env",
        type=str,
        choices=list(wappstoUrl.keys()),
        default="prod",
        help="Wappsto environment."
    )
    parser.add_argument(
        "--token",
        type=uuid.UUID,
        help="The Session Token. If not given, you are prompted to login."
    )
    parser.add_argument(
        "--path",
        type=pathlib.Path,
        default=".",
        help="The location to which the config files are saved."
    )
    parser.add_argument(
        "--recreate",
        type=uuid.UUID,
        help="Recreate Config file, for given network UUID. (Overwrites existent)"
    )
    parser.add_argument(
        "--debug",
        action='store_true',
        help="Make the operation more talkative",
    )

    args = parser.parse_args(sys.argv[1:])

    debug = args.debug if args.debug else False

    base_url = wappstoUrl[args.env]

    if not args.token:
        session = _start_session(
            base_url=base_url,
            username=input("Wappsto Username: "),
            password=getpass.getpass(prompt="Wappsto Password: "),
        )
    else:
        session = args.token
    if not args.recreate:
        creator = _create_network(
            session=session,
            base_url=base_url,
            dry_run=args.dry_run
        )
        _claim_network(
            session=session,
            base_url=base_url,
            network_uuid=creator.get('network', {}).get('id'),
            dry_run=args.dry_run
        )
    else:
        creator = _get_network(
            session=session,
            base_url=base_url,
            network_uuid=args.recreate,
        )

    args.path.mkdir(exist_ok=True)

    _create_certificaties_files(args.path, creator, args.dry_run)

    print("\nEnjoy...")
    exit(0)


if __name__ == "__main__":
    main()



from . import Timestamp as Timestamp
from _typeshed import Incomplete
from enum import Enum
from typing import List, Optional

force_trace: bool

class TraceStatus(str, Enum):
    PENDING: str
    OK: str
    FAIL: str

class TraceWrapper:
    log: Incomplete
    class_obj: Incomplete
    tracer: Incomplete
    tracing_obj: Incomplete
    def __init__(self, class_obj: object, tracer: object, tracing_obj: List[str]) -> None: ...
    def __getattr__(self, attr_name): ...

class Trace:
    log: Incomplete
    id: Incomplete
    parent: Incomplete
    name: Incomplete
    timestamp: Incomplete
    def __init__(self, name: str, id: Optional[str] = ..., parent: Optional[str] = ...) -> None: ...
    @property
    def status(self) -> str: ...
    @staticmethod
    def trace_check(jsonrpc_elemt: dict, name: str) -> Optional['Trace']: ...
    @staticmethod
    def trace_list_check(jsonrpc_elemts: list, name: str) -> Optional[List['Trace']]: ...
    def tracing(self, name: Optional[str] = ...): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def get_child(self) -> Trace: ...
    def send_pending(self, name: Optional[str] = ...): ...
    def send_ok(self, name: Optional[str] = ...): ...
    def send_failed(self, name: Optional[str] = ...): ...



import uuid
from ..schema.base_schema import PermissionType as PermissionType, timestamp_converter as timestamp_converter
from ..schema.iot_schema import WappstoMethod as WappstoMethod
from ..service.template import ServiceClass as ServiceClass
from ..utils.Timestamp import str_to_datetime as str_to_datetime
from .device import Device as Device
from .template import ValueBaseType as ValueBaseType
from _typeshed import Incomplete
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Optional, Union

class Period(str, Enum):
    PERIODIC_REFRESH: str
    DROP_UNTIL: str

class Delta(str, Enum):
    ONLY_UPDATE_IF: str
    EXTRA_UPDATES: str

class Value:
    class RequestType(str, Enum):
        refresh: str
        control: str
        delete: str
    class ChangeType(str, Enum):
        report: str
        delta: str
        period: str
        name: str
        description: str
        unit: str
        min: str
        max: str
        step: str
        encoding: str
        meaningful_zero: str
        state: str
    log: Incomplete
    value_type: Incomplete
    schema: Incomplete
    report_state: Incomplete
    control_state: Incomplete
    parent: Incomplete
    element: Incomplete
    children_name_mapping: Incomplete
    connection: Incomplete
    def __init__(self, parent: Device, value_type: ValueBaseType, name: str, value_uuid: Optional[uuid.UUID], type: Optional[str] = ..., permission: PermissionType = ..., min: Optional[Union[int, float]] = ..., max: Optional[Union[int, float]] = ..., step: Optional[Union[int, float]] = ..., encoding: Optional[str] = ..., xsd: Optional[str] = ..., namespace: Optional[str] = ..., period: Optional[int] = ..., delta: Optional[Union[int, float]] = ..., description: Optional[str] = ..., meaningful_zero: Optional[str] = ..., mapping: Optional[bool] = ..., ordered_mapping: Optional[bool] = ..., si_conversion: Optional[str] = ..., unit: Optional[str] = ...) -> None: ...
    def getControlData(self) -> Optional[Union[float, str]]: ...
    def getControlTimestamp(self) -> Optional[datetime]: ...
    def getReportData(self) -> Optional[Union[float, str]]: ...
    def getReportTimestamp(self) -> Optional[datetime]: ...
    @property
    def name(self) -> str: ...
    @property
    def uuid(self) -> uuid.UUID: ...
    def onChange(self, callback: Callable[[Value], None]) -> Callable[[Value], None]: ...
    def cancelOnChange(self) -> None: ...
    def onReport(self, callback: Callable[[Value, Union[str, float]], None]) -> Callable[[Value, Union[str, float]], None]: ...
    def cancelOnReport(self) -> None: ...
    def onControl(self, callback: Callable[[Value, Union[str, float]], None]) -> Callable[[Value, Union[str, float]], None]: ...
    def cancelOnControl(self) -> None: ...
    def onCreate(self, callback: Callable[[Value], None]) -> Callable[[Value], None]: ...
    def cancelOnCreate(self) -> None: ...
    def onRefresh(self, callback: Callable[[Value], None]) -> Callable[[Value], None]: ...
    def cancelOnRefresh(self) -> None: ...
    def onDelete(self, callback: Callable[[Value], None]) -> Callable[[Value], None]: ...
    def cancelOnDelete(self) -> None: ...
    def refresh(self) -> None: ...
    def change(self, name: str, value: Any) -> None: ...
    def delete(self) -> None: ...
    def report(self, value: Union[int, float, str, None], timestamp: Optional[datetime] = ...) -> None: ...
    def control(self, value: Union[int, float, str, None], timestamp: Optional[datetime] = ...) -> None: ...
    def close(self) -> None: ...



from ..utils import observer as observer
from .protocol import Connection as Connection, StatusID as StatusID
from _typeshed import Incomplete
from pathlib import Path
from typing import Any, Callable, Optional, Union

class TlsSocket(Connection):
    log: Incomplete
    observer_name: str
    observer: Incomplete
    send_ready: Incomplete
    address: Incomplete
    port: Incomplete
    socket_timeout: int
    RECEIVE_SIZE: int
    killed: Incomplete
    ssl_context: Incomplete
    def __init__(self, address: str, port: int, ca: Path, crt: Path, key: Path) -> None: ...
    def send(self, data: Union[str, bytes]) -> bool: ...
    def receive(self, parser: Callable[[bytes], Any]) -> Any: ...
    def connect(self) -> Optional[bool]: ...
    def reconnect(self, retry_limit: Optional[int] = ...) -> bool: ...
    def disconnect(self) -> None: ...
    socket: Incomplete
    raw_socket: Incomplete
    def close(self) -> None: ...



"""
"""
import logging
import time

from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import Path


from typing import List
from typing import Optional
from typing import Union


class OfflineStorage(ABC):
    @abstractmethod
    def save(self, data: str) -> None: ...
    @abstractmethod
    def load(self, max_count: Optional[int] = None) -> List[str]: ...


class OfflineStorageFiles(OfflineStorage):
    """
    Timed over 1M runs.

    In [5]: kk = write()
    Time used per run: 40_965.816834ns

    In [7]: kk = read()
    Time used per run: 173_536.595006ns
    """
    def __init__(self, location: Union[Path, str]):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        if not isinstance(location, Path):
            location = Path(location)

        if location.is_file():
            raise NotADirectoryError(
                f"The given location need to be a directory, not: {location}"
            )
        self.loc = location

        self.suffix = ".data"
        self.loc.mkdir(exist_ok=True)
        self.log.info(f"Location created: {self.loc}")
        self._files = sorted(
            x for x in self.loc.iterdir()
            if x.is_file() and x.suffixes and x.suffixes[-1] == self.suffix
        )

    def _sort_files(self, count) -> List[Path]:
        temp = self._files[:count]
        del self._files[:count]
        return temp

    @contextmanager
    def auto_save(self, data: str):
        """Used when replying on a trace."""
        # self.send_pending(name=name)  # Are send on class creation.
        try:
            yield
        except Exception:
            self.save(data=data)
            raise

    def save(self, data: str) -> None:
        self.log.debug(f"Saving data: {data}")
        datafile = self.loc / (str(time.perf_counter_ns()) + self.suffix)
        self.log.debug(f"Save to file: {datafile}")
        with datafile.open(mode="w") as file:
            file.write(f"{data}")
        self._files.append(datafile)

    def load(self, max_count: Optional[int] = None) -> List[str]:
        self.log.debug(f"Load {max_count} lines.")
        data_files = self._sort_files(max_count)
        data = []
        for data_file in data_files:
            with data_file.open(mode="r") as file:
                data.append(file.read())
            self.log.debug(f"File Loaded: {data_file}")
            data_file.unlink()
        return data



from enum import Enum
from pydantic import BaseModel
from typing import Dict, Optional

class IoTEvent(str, Enum):
    CREATE: str
    CHANGE: str
    REQUEST: str
    DELETE: str

class ValueBaseType(str, Enum):
    STRING: str
    NUMBER: str
    BLOB: str
    XML: str

class ValueSettinsSchema(BaseModel):
    value_type: ValueBaseType
    type: str
    mapping: Optional[Dict]
    ordered_mapping: Optional[bool]
    meaningful_zero: Optional[bool]
    si_conversion: Optional[str]
    min: Optional[float]
    max: Optional[float]
    step: Optional[float]
    encoding: Optional[str]
    xsd: Optional[str]
    namespace: Optional[str]
    unit: Optional[str]

class ValueTemplate(str, Enum):
    ADDRESS_NAME: str
    ALTITUDE_M: str
    ANGLE: str
    APPARENT_POWER_VA: str
    BLOB: str
    BOOLEAN_ONOFF: str
    BOOLEAN_TRUEFALSE: str
    CITY: str
    CO2_PPM: str
    COLOR_HEX: str
    COLOR_INT: str
    COLOR_TEMPERATURE: str
    CONCENTRATION_PPM: str
    CONNECTION_STATUS: str
    COUNT: str
    COUNTRY: str
    COUNTRY_CODE: str
    CURRENT_A: str
    DISTANCE_M: str
    DURATION_MIN: str
    DURATION_MSEC: str
    DURATION_SEC: str
    EMAIL: str
    ENERGY_KWH: str
    ENERGY_MWH: str
    ENERGY_WH: str
    FREQUENCY_HZ: str
    HUMIDITY: str
    IDENTIFIER: str
    IMAGE_JPG: str
    IMAGE_PNG: str
    IMPULSE_KWH: str
    INTEGER: str
    JSON: str
    LATITUDE: str
    LONGITUDE: str
    LUMINOUSITY_LX: str
    NUMBER: str
    ORGANISATION: str
    PERCENTAGE: str
    PHONE: str
    POSTCODE: str
    POWER_KW: str
    POWER_WATT: str
    PRECIPITATION_MM: str
    PRESSURE_HPA: str
    REACTIVE_ENERGY_KVARH: str
    REACTIVE_POWER_KVAR: str
    SPEED_MS: str
    STREET: str
    STRING: str
    TEMPERATURE_CELSIUS: str
    TEMPERATURE_FAHRENHEIT: str
    TEMPERATURE_KELVIN: str
    TIMESTAMP: str
    UNIT_TIME: str
    VOLTAGE_V: str
    VOLUME_M3: str
    XML: str

valueSettings: Dict[ValueTemplate, ValueSettinsSchema]



from .connections import protocol as connection
from .modules.device import Device as Device
from .modules.network import Network as Network
from .modules.template import ValueTemplate as ValueTemplate
from .modules.value import PermissionType as PermissionType, Value as Value
from .service import template as service
from .utils.offline_storage import OfflineStorage as OfflineStorage
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Optional, Union

def onStatusChange(StatusID: Union[service.StatusID, connection.StatusID], callback: Callable[[Union[service.StatusID, connection.StatusID], Any], None]): ...

class ConnectionTypes(str, Enum):
    IOTAPI: str
    RESTAPI: str

def config(config_folder: Union[Path, str] = ..., connection: ConnectionTypes = ..., fast_send: bool = ..., ping_pong_period_sec: Optional[int] = ..., offline_storage: Union[OfflineStorage, bool] = ...) -> None: ...
def createNetwork(name: str, description: str = ...) -> Network: ...
def connect() -> None: ...
def disconnect() -> None: ...
def close() -> None: ...



from _typeshed import Incomplete
from typing import Any, Callable, Dict, List

obs_log: Incomplete
default_subscriber: Incomplete
subscriber: Dict[str, List[Callable[[str, Any], None]]]

def subscribe(event_name: str, callback: Callable[[str, Any], None]) -> None: ...
def post(event_name: str, data: Any) -> None: ...
def unsubscribe(event_name: str, callback: Callable[[str, Any], None]) -> bool: ...



import logging
import random
import string

from contextlib import contextmanager
from enum import Enum

from typing import List
from typing import Optional


import requests

from . import Timestamp


force_trace: bool = False


class TraceStatus(str, Enum):
    PENDING = "pending"  # On Receive
    OK = "ok"            # If Send Successful
    FAIL = "fail"        # If Send Failed.


class TraceWrapper:
    """."""

    def __init__(
        self,
        class_obj: object,
        tracer: object,
        tracing_obj: List[str]
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())
        self.class_obj = class_obj
        self.tracer = tracer
        self.tracing_obj = tracing_obj

    def __getattr__(self, attr_name):
        if attr_name not in self.tracing_obj:
            return getattr(self.class_obj, attr_name)

        def wrapper(*args, **kwargs):
            with self.tracer.tracing():
                getattr(self.class_obj, attr_name)(*args, **kwargs)

        return wrapper


class Trace:
    def __init__(
        self,
        name: str,
        id: Optional[str] = None,
        parent: Optional[str] = None
    ):

        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.id = id if id else Trace._generateId()
        self.parent = parent
        self.name = name
        self.__status: TraceStatus = TraceStatus.PENDING
        self.timestamp = Timestamp.timestamp()

    @property
    def status(self) -> str:
        """Returns the Status of the Trace."""
        return self.__status

    @staticmethod
    def _generateId() -> str:
        """Generate a Trace ID."""
        return "WappstoIoT_" + "".join(random.choices(
            string.ascii_letters + string.digits,
            k=10
        ))

    @staticmethod
    def _find_parent_id(jsonrpc_elemt) -> Optional[str]:
        """
        Check if a trace package should be send if so it returns the parent ID.
        """
        if isinstance(jsonrpc_elemt, dict):
            return jsonrpc_elemt.get(
                'params', {}).get(
                'meta', {}).get(
                'trace', None
            )
        return jsonrpc_elemt.params.meta.trace

    @staticmethod
    def trace_check(jsonrpc_elemt: dict, name: str) -> Optional['Trace']:
        global force_trace
        parent = Trace._find_parent_id(jsonrpc_elemt)
        if parent or force_trace:
            t = Trace(
                parent=parent,
                name=name
            )
            t.send_pending(name=name)
            return t
        return None

    @staticmethod
    def trace_list_check(
        jsonrpc_elemts: list,
        name: str
    ) -> Optional[List['Trace']]:
        r_list = []
        for t in jsonrpc_elemts:
            t = Trace.trace_check(t, name=name)
            if t:
                r_list.append(t)
        return r_list

    @contextmanager
    def tracing(self, name: Optional[str] = None):
        """Used when replying on a trace."""
        # self.send_pending(name=name)  # Are send on class creation.
        try:
            yield
        except Exception:
            self.send_failed(name=name)
            raise
        else:
            self.send_ok(name=name)

    def __enter__(self):
        """Used when starting a new trace."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.send_failed()
        else:
            self.send_ok()

    def __repr__(self) -> str:
        return (
            "Trace("
            f"ID={self.id}; "
            f"Parent={self.parent}; "
            f"Status={self.__status})"
        )

    def get_child(self) -> 'Trace':
        return Trace(
            parent=self.id,
            name=f"{self.name} child",
        )

    def send_pending(self, name: Optional[str] = None):
        self._send(name=name, status=TraceStatus.PENDING)

    def send_ok(self, name: Optional[str] = None):
        self._send(name=name, status=TraceStatus.OK)

    def send_failed(self, name: Optional[str] = None):
        self._send(name=name, status=TraceStatus.FAIL)

    def _send(
        self,
        status: TraceStatus,
        name: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Send a Package Trace to Seluxit.

        Package tracing are used to debug, where the given package are lost,
        and/or the timing of the given package through the system.

        A trace package should be send when a JSONRPC object, with a trace value
        in the params, meta field, are received. The trace package should contain
        a status that are set to 'pending'.
        When the reply for the traced Wappsto json are ready to be send
        (to the socket). The Wappsto json's meta field should have the trace filed
        added, with the trace_id & name the pending trace package was send with,
        followed with the sending of another trace-package, where the status 'ok'.
        If for some reason it was not possible to generate a reply (but there
        should have been), the trace package should then be send with the
        'fail' status, on the realization this is the case.

        Args:
            trace_id: A generated ID, that should be added to
                      the Wappsto json meta trace filed.
            parent_id: The trace-value from the Wappsto json meta field.
            name: A descriptive name.
            status: Status for the traced package.
            timestamp: the timestamp in the ISO format.
                       If not sat, it will be sat in the time of sending.

        Returns:
            True, of the Trace package was send successful,
            False, if it was not.
        """
        self.__status = status
        params = {
            "id": self.id,
            "name": name if name else self.name,
            "status": status,
            "timestamp": self.timestamp
        }
        if self.parent:
            params["parent"] = self.parent

        self.log.info(f"Trace id: {self.id}")

        r_data = requests.post(
            url='https://tracer.iot.seluxit.com/trace',
            params=params
        )
        self.log.debug(f"Trace reply: {r_data.text}")

        if r_data.status_code >= 300:
            self.log.error(f"Trace http error code: {r_data.status_code}")
            return None

        self.log.debug("Trace send!")

        return params



"""Contain name checking functions."""
import string


wappsto_letters = string.digits + string.ascii_letters + " -_.~"
__wappsto_letter_set = set(wappsto_letters)


def legal_name(name: str) -> bool:
    """
    Check if the given name is legal.

    Args:
        name: the name to check.
    Return:
        True, if it is legal,
        False, if it is illegal.
    """
    return not set(name) - __wappsto_letter_set


def illegal_characters(name: str) -> str:
    """
    Check if the given name is illegal.

    Args:
        name: the name to check.
    Return:
        string with all the illegal characters in the name.
    """
    return ''.join(set(name) - __wappsto_letter_set)



"""
Contain the Observer functionality.

The Observer Pattern, is as design pattern, that 
"""
import logging
import threading

from typing import Any
from typing import Dict
from typing import List
from typing import Callable


obs_log = logging.getLogger(__name__)
obs_log.addHandler(logging.NullHandler())

default_subscriber = [
    lambda event_name, data: obs_log.debug(
        f"Default Observer: event_name={event_name}, data={data}"
    )
]

subscriber: Dict[str, List[Callable[[str, Any], None]]] = {}


def subscribe(event_name: str, callback: Callable[[str, Any], None]) -> None:
    """
    Subscribe to the event with given event name.

    Note: If lambda was used to subscribe with, it need to be the same
    instance that is used to unsubscribe with. So if it is needed to
    unsubscribe, save the instance.

    Args:
        event_name: The Unique name for the wanted event.
        callback: The function that need triggered on the given event.
            The function will be called with the 'event_name', and 'data',
            that the event generate.
    """
    obs_log.debug(f"New Subscriber: {event_name}, {callback}")
    if event_name not in subscriber:
        subscriber[event_name] = default_subscriber.copy()
    subscriber[event_name].append(callback)


def post(event_name: str, data: Any) -> None:
    """
    Post the event with given event name.

    Args:
        event_name: An unique name for the given event.
        data: The given event subscriber might want.
    """
    def executer():
        for fn in subscriber.get(event_name, default_subscriber):
            fn(event_name, data)

    # NOTE: Needed for ensure that the receive thread do not get blocked.
    th = threading.Thread(target=executer)
    th.start()


def unsubscribe(event_name: str, callback: Callable[[str, Any], None]) -> bool:
    """
    Unsubscribe from given event name.

    Note: if lambda was used to subscribe with, it need to be the same
    instance that is used to unsubscribe with.

    Args:
        event_name: The Unique name for the wanted event.
        callback: The function that need triggered on the given event.
            The function will be called with the 'event_name', and 'data',
            that the event generate.

    Returns:
        True, is the function was removed from the subscriber list.
        False, if it could not be, as in could not find it.
    """
    obs_log.debug(f"Unsubscribing: {event_name}, {callback}")

    try:
        # NOTE: Faster then testing if callback is in the list, when it is (60%+).
        subscriber.get(event_name, []).remove(callback)
    except ValueError:
        return False
    else:
        return True



from ..connections.protocol import Connection as Connection
from ..connections.sslsocket import TlsSocket as TlsSocket
from ..schema.base_schema import BlobValue as BlobValue, Device as Device, IdList as IdList, Network as Network, NumberValue as NumberValue, State as State, StringValue as StringValue, WappstoObject as WappstoObject, XmlValue as XmlValue
from ..schema.iot_schema import Identifier as Identifier, JsonData as JsonData, JsonReply as JsonReply, Success as Success, WappstoMethod as WappstoMethod
from ..utils import observer as observer, tracer as tracer
from ..utils.certificateread import certificate_info_extraction as certificate_info_extraction
from .template import ServiceClass as ServiceClass, StatusID as StatusID
from _typeshed import Incomplete
from pathlib import Path
from typing import Callable, Optional, Union
from uuid import UUID

ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

class IoTAPI(ServiceClass):
    wappstoPort: Incomplete
    log: Incomplete
    ca: Incomplete
    crt: Incomplete
    key: Incomplete
    timeout: int
    fast_send: Incomplete
    connection: Incomplete
    subscribers: Incomplete
    jsonrpc: Incomplete
    killed: Incomplete
    workers: Incomplete
    def __init__(self, ca: Path, crt: Path, key: Path, worker_count: int = ..., fast_send: bool = ...) -> None: ...
    def close(self) -> None: ...
    def ping(self): ...
    def subscribe_network_event(self, uuid: UUID, callback: Callable[[Network, WappstoMethod], None]): ...
    def unsubscribe_network_event(self, uuid: UUID, callback: Callable[[Network, WappstoMethod], None]): ...
    def post_network(self, data: Network) -> bool: ...
    def put_network(self, uuid: UUID, data: Network) -> bool: ...
    def get_network(self, uuid: UUID) -> Optional[Network]: ...
    def delete_network(self, uuid: UUID) -> bool: ...
    def subscribe_device_event(self, uuid: UUID, callback: Callable[[Device, WappstoMethod], None]): ...
    def unsubscribe_device_event(self, uuid: UUID, callback: Callable[[Device, WappstoMethod], None]): ...
    def post_device(self, network_uuid: UUID, data: Device) -> bool: ...
    def put_device(self, uuid: UUID, data: Device) -> bool: ...
    def get_device_where(self, network_uuid: UUID, **kwargs: str) -> Optional[UUID]: ...
    def get_device(self, uuid: UUID) -> Union[Device, None]: ...
    def delete_device(self, uuid: UUID) -> bool: ...
    def subscribe_value_event(self, uuid: UUID, callback: Callable[[ValueUnion, WappstoMethod], None]): ...
    def unsubscribe_value_event(self, uuid: UUID, callback: Callable[[Device, WappstoMethod], None]): ...
    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool: ...
    def put_value(self, uuid: UUID, data: ValueUnion) -> bool: ...
    def get_value_where(self, device_uuid: UUID, **kwargs: str) -> Optional[UUID]: ...
    def get_value(self, uuid: UUID) -> Union[ValueUnion, None]: ...
    def delete_value(self, uuid: UUID) -> bool: ...
    def subscribe_state_event(self, uuid: UUID, callback: Callable[[State, WappstoMethod], None]): ...
    def unsubscribe_state_event(self, uuid: UUID, callback: Callable[[Device, WappstoMethod], None]): ...
    def post_state(self, value_uuid: UUID, data: State) -> bool: ...
    def put_state(self, uuid: UUID, data: State) -> bool: ...
    def get_state(self, uuid: UUID) -> Union[State, None]: ...
    def delete_state(self, uuid: UUID) -> bool: ...
